use std::{
    iter::{self, from_fn},
    num::ParseIntError,
    path::Path,
};

// TODO: tree_sitter ?

#[derive(Debug, PartialEq)]
pub enum Token {
    Integer(i64),
    Operator(char),
    OpenParen,
    CloseParen,
}

#[derive(Debug, PartialEq, Clone, Copy)]
pub enum Source<'a> {
    File(&'a Path),
    Unknown,
}

#[derive(Debug, PartialEq)]
pub struct Position<'a> {
    source: Source<'a>, // TODO: not certain how this will work in the real world, but I don't want all these positions taking up a bunch of memory unnecessarily as we lex and parse, so for now we deal with lifetimes so we can store references and have Source be cheap to clone everywhere
    line: usize,
    col: usize,
}

#[derive(Debug, PartialEq)]
pub enum Error {
    // TODO: not a number error?
    ParseIntError(ParseIntError),
    Unexpected(char),
}

pub struct Lexer {
    num_tokens: usize,
}

impl Default for Lexer {
    fn default() -> Self {
        Self { num_tokens: 0 }
    }
}

impl Lexer {
    pub fn num_tokens(&self) -> usize {
        self.num_tokens
    }

    pub fn lex_str<'a>(
        &mut self,
        code: &'a str,
    ) -> Result<Vec<(Position<'a>, Token)>, (Position<'a>, Error)> {
        self.lex(code, None)
    }

    pub fn lex<'a>(
        &mut self,
        code: &'a str,
        source: Option<Source<'a>>,
    ) -> Result<Vec<(Position<'a>, Token)>, (Position<'a>, Error)> {
        let source: Source = source.unwrap_or(Source::Unknown);
        let mut result = vec![];
        let mut scanner = Self::walk(code).peekable();
        while let Some((line, col, c)) = scanner.next() {
            let pos = move || Position { line, col, source };
            let err = |err: Error| Err((pos(), err));
            let token: Token = match c {
                '(' => Token::OpenParen,
                ')' => Token::CloseParen,
                '\t' | ' ' => {
                    continue;
                }
                '0'..='9' => {
                    let s = iter::once(c)
                        .chain(from_fn(|| {
                            scanner
                                .by_ref()
                                .next_if(|(_, _, c)| c.is_ascii_digit())
                                .map(|(_, _, c)| c)
                        }))
                        .collect::<String>();
                    let result = match s.parse() {
                        Ok(i) => Token::Integer(i),
                        Err(e) => return err(Error::ParseIntError(e)),
                    };
                    result
                }
                '+' | '/' | '%' | '*' | '-' => Token::Operator(c),
                c => return err(Error::Unexpected(c)),
            };
            result.push((pos(), token));
            // an excuse for mutability at present
            self.num_tokens += 1
        }
        return Ok(result);
    }

    fn walk(code: &str) -> impl Iterator<Item = (usize, usize, char)> + use<'_> {
        Self::lines(code)
            .map(|(line_num, line)| {
                Self::cols(line).map(move |(col_num, c)| (line_num, col_num, c))
            })
            .flatten()
    }

    fn lines<'a>(code: &'a str) -> impl Iterator<Item = (usize, &'a str)> {
        code.lines().enumerate().map(|(i, line)| (i + 1, line))
    }

    fn cols(line: &str) -> impl Iterator<Item = (usize, char)> + use<'_> {
        line.chars().enumerate().map(|(i, c)| (i + 1, c))
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn addition_operation() {
        let pos = |col| Position {
            line: 1,
            col,
            source: Source::Unknown,
        };
        let mut lexer = Lexer::default();
        assert_eq!(
            lexer.lex_str("3 + 9"),
            Ok(vec![
                (pos(1), Token::Integer(3)),
                (pos(3), Token::Operator(ast::Operator::Add)),
                (pos(5), Token::Integer(9)),
            ])
        );
        assert_eq!(lexer.num_tokens(), 3);
    }
}
