use crate::{ast::*, lexer};

struct Parser {
    num_tokens_parsed: usize,
}

impl Default for Parser {
    fn default() -> Self {
        Self {
            num_tokens_parsed: 0,
        }
    }
}

#[derive(Debug, PartialEq)]
pub enum Error {
    Lexer((lexer::Position<'static>, lexer::Error)),
    Unexpected((lexer::Position<'static>, lexer::Token)),
}

impl Parser {
    pub fn parse_str(&mut self, code: &'static str) -> Result<Expression, Error> {
        let mut lexer = lexer::Lexer::default();
        let tokens = lexer.lex_str(code).map_err(|e| Error::Lexer(e))?;

        Ok(self.parse(tokens)?)
    }

    pub fn parse(
        &mut self,
        tokens: Vec<(lexer::Position, lexer::Token)>,
    ) -> Result<Expression, Error> {
        let mut iter = tokens.iter().peekable();
        while let Some((_p, token)) = iter.next() {
            match iter.peek() {
                Some(_) => Expression::Infix,
                _ => {}
            }
            self.num_tokens_parsed += 1
        }
        Ok(ast::Expression::Unit)
    }

    pub fn num_tokens_parsed(&self) -> usize {
        self.num_tokens_parsed
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use ast::*;

    #[test]
    fn addition_operation() {
        let mut parser = Parser::default();
        assert_eq!(
            parser.parse_str("3 + 9"),
            Ok(Expression::InfixOperation(
                Box::new(Expression::Integer(3)),
                Operator::Add,
                Box::new(Expression::Integer(9))
            ))
        );
        assert_eq!(parser.num_tokens_parsed(), 3);
    }
}
